# Copyright 2017, 2018 CA Technologies. All Rights Reserved.

import os,sys,platform
import re
import getpass
import subprocess
try:
    from pwd import getpwuid
except:
    pass


class EnvironmentalKey:
    def __init__(self, konf, logger):
        self.konf = konf
        self.log  = logger
        # conf_file_name, conf_path, ac_server, ac_workspace, ac_project, gh_server
        self.ac_server = konf.topLevel('AgileCentral').get('Server', 'rally1.rallydev.com')

    def executeShellCommand(self, cmd, pattern=None, dosplit=True, re_flags=''):
        self.log.debug("Executing command: %s" % cmd)

        pipe = None
        try:
            pipe = subprocess.Popen(cmd, shell=True, cwd='.', stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        except:
            ex_type, ex_instance, ex_tb = sys.exc_info()
            self.log.error('unable to Popen for cmd: %s, Exception(%s) info: %s' % (cmd, ex_type, ex_instance.value.args[0]))
            return None
        self.log.debug("  command execution returned a pipe? %s" % (pipe is not None))

        out, error = None, None
        try:
            out, error = pipe.communicate()
        except:
            ex_type, ex_instance, ex_tb = sys.exc_info()
            self.log.error('unable to pipe.communicate for cmd: %s, Exception(%s) info: %s' % (cmd, ex_type, ex_instance.value.args[0]))
            return None
        self.log.debug("  command execution pipe.communicated returned out? %s  and error? %s" % (out is not None, error is not None))

        rc = None
        try:
            rc = pipe.wait()
        except:
            ex_type, ex_instance, ex_tb = sys.exc_info()
            self.log.error('unable to pipe.wait for cmd: %s, Exception(%s) info: %s' % (cmd, ex_type, ex_instance.value.args[0]))
            return None

        self.log.debug("  pipe.wait return code of %s" % (rc))

        command_output = ""
        try:
            command_output = out.decode()
            try:
                self.log.debug('attempt to decode the pipe out content returned:\n%s' % command_output)
            except:
                ex_type, ex_instance, ex_tb = sys.exc_info()
                self.log.debug('unable to show content of pipe out content, Exception(%s) info: %s' % (ex_type, ex_instance.value.args[0]))
        except:
            ex_type, ex_instance, ex_tb = sys.exc_info()
            self.log.error("Unable to decode output from command: %s  Exception(%s) info: %s" % (cmd, ex_type, ex_instance.value.args[0]))
            return None

        if not pattern:  # if there is no pattern to match, then send back the command_output intact (if dosplit then split it into discrete lines)
            if dosplit:
                command_output = command_output.split("\n")
            return command_output

        if dosplit:
            self.log.debug("  the command output is to be split on newline chars...")
            output_lines = ""

            try:
                output_lines = command_output.split("\n")
            except:
                ex_type, ex_instance, ex_tb = sys.exc_info()
                self.log.debug('unable to split command output, Exception(%s) info: %s' % (ex_type, ex_instance.value.args[0]))

            lines = [line.strip() for line in output_lines]
            output_lines = [line for line in lines if re.search(pattern, line, re.I | re.M)]
            self.log.debug(" result of pattern match on lines is:\n%s" % repr(output_lines))
            return output_lines
        else:
            self.log.debug("  the command output is NOT to be split on newline chars (or there aren't expected to be any newlines in the output)")
            result = ""
            try:
                self.log.debug('   looking for pattern: %s in the output string (shown above)' % pattern)
                result = re.search(pattern, command_output, re.I | re.M)
                if result:
                    self.log.debug("  result of pattern match on string is: %s" % repr(result.groups(0)))
                else:
                    self.log.debug("  result of pattern match on string is: None")
                return result
            except:
                ex_type, ex_instance, ex_tb = sys.exc_info()
                self.log.error('unable to detect regex pattern in output, Exception(%s) info: %s' % (ex_type, ex_instance.value.args[0]))
            return result


    def getMacAddress(self, operating_system):
        macaddr = 'mac address not obtained'
        command = 'ifconfig -a'
        self.log.debug("Detected operating system %s" % operating_system)
        if operating_system == 'mac':
            pattern = r'(e[a-z][a-z0-9]*\d+): .*?\n\s+ether (([a-f0-9]{2}:){5}[a-f0-9]{2})\s*\n'
            # ifc, macaddr = pattern.match_object.group 0 and 1
        elif operating_system == 'posix':
            pattern = r'(e[a-z][a-z0-9]*\d+)\s+Link encap:Ethernet \s*HWaddr (([a-f0-9]{2}:){5}[a-f0-9]{2})\s*\n'
            # ifc, macaddr = pattern.match_object.group 0 and 1
        elif operating_system == 'windows':
            command = 'ipconfig /all'
            pattern = r'Physical Address.*?(([A-F0-9]{2}-){5}[A-F0-9]{2})'
            # macaccr = pattern.match_object.group 0
        else:
            return '89:AF:43:BC:71:DD'

        result = self.executeShellCommand(command, pattern, dosplit=False, re_flags='im')

        if operating_system in ['mac', 'posix']:
            try:
                macaddr = result.group(2)
            except:
                command = 'ip link'
                pattern = r'\s+\w+\/ether\s+([0-9A-Z:]+)\s'
                result = self.executeShellCommand(command, pattern, dosplit=False, re_flags='im')
                self.log.debug("getMacAddress result executing ip link for operating system %s: %s" % (operating_system, repr(result.groups())))
                if result and result.groups():
                    macaddr = result.group(1)
                else:
                    macaddr =  '89:AF:BD:43:71:DD'
        else:
            try:
                macaddr = result.group(0).split(':')[1]
            except:
                ex_type, ex_instance, ex_tb = sys.exc_info()
                macaddr = 'DD:AF:43:BC:71:89'
                self.log.debug("Acquired a mac address value...")
        return macaddr

    def getHardwareId(self, operating_system):
        # if operating_system is 'mac' then we can get both hardware_uuid and serial_number (combine them)
        if operating_system == 'mac':
            hardware_uuid, serial_number = '', ''
            cmd    = "system_profiler SPHardwareDataType"
            result = self.executeShellCommand(cmd, pattern='(Serial Number|Hardware UUID)')

            for line in result:
                line = line.lstrip()
                if line.startswith('Serial Number (system):'):
                    serial_number = line.split(': ')[1]
                elif line.startswith('Hardware UUID:'):
                    hardware_uuid = line.split(': ')[1]
            return "-".join([hardware_uuid, serial_number])

        elif operating_system == 'windows':
            pairs = [('Board:', 'baseboard'), ('BIOS:', 'bios'), ('OSL:', 'os')]

            values = []
            for label, target in pairs:
                cmd = "wmic %s get serialnumber" % target
                result = self.executeShellCommand(cmd, dosplit=False)
                value = result.replace('SerialNumber', '').strip()
                values.append('%s%s' % (label, value))
            return '-'.join(values)

        elif operating_system == 'posix':
            values = []
            cmd = "cat /etc/fstab"
            result = self.executeShellCommand(cmd, pattern='UUID=[a-f0-9-]+ /(boot)? ')
            for line in result:
                truncated_line = line.split()[0].replace('UUID=','')
                values.append(truncated_line)
            return '-'.join(values)
        else:
            return "It is hard to make predictions, especially about the future"

    def getOwner(self, operating_system, config=None):
        owner = 'Yogi Berra'
        if operating_system == 'mac' or operating_system == 'posix':
            if not config:
                owner = getpwuid(os.stat(os.getcwd()).st_uid).pw_name
            else:
                owner = getpwuid(os.stat(config).st_uid).pw_name
        elif operating_system == 'windows':
            if not config:
                cmd = "dir /q"
                result = self.executeShellCommand(cmd)
                match = "<DIR>"
                lines = [line for line in result if match in line]
                if len(lines) > 0:
                    first_line = lines[0]
                    owner = first_line.split('>')[1].split()[0]
            else:
                config_name = os.path.basename(config)
                cmd = "dir /q %s" % os.path.normpath(config)
                result = self.executeShellCommand(cmd)
                lines = [line for line in result if config_name in line]
                if len(lines) > 0:
                    first_line = lines[0]
                    try:
                        owner = first_line.rsplit(None, 1)[0].rsplit(None, 1)[-1]
                    except:
                        self.log.debug("Acquired a config file owner value ...")

        return owner

    def osFamily(self):
        name = platform.system()
        return {'Darwin' : 'mac',
                'Linux'  : 'posix',
                'Windows': 'windows'
               }[name]

    def identVector(self):
        operating_system = None
        try:
            operating_system = self.osFamily()
            self.log.debug('Obtained OS family as: %s' % operating_system)
        except:
            self.log.error('Unable to obtain OS family, %s' % sys.exc_info()[0])
            operating_system = 'UnidentifiedOSFamily'

        config_path = self.konf.config_file_path
        ac_conf     = self.konf.topLevel('AgileCentral')

        hostname = None
        try:
            hostname      = self.executeShellCommand('hostname')[0].strip()
            self.log.debug('Obtained hostname as: %s' % hostname)
        except:
            self.log.error('Unable to obtain hostname, %s' % sys.exc_info()[0])
            hostname      = 'UnidentifiedHost'

        hardware_id = None
        try:
            hardware_id = self.getHardwareId(operating_system)
            self.log.debug('Obtained hardware_id as: %s' % hardware_id)
        except:
            self.log.error('Unable to obtain hardware_id, %s' % sys.exc_info()[0])
            hardware_id = 'UnobtainableHardwareID'

        mac_address = None
        try:
            mac_address = self.getMacAddress(operating_system)
            self.log.debug('Obtained mac address as: %s' % mac_address)
        except:
            exc_type, exc_value, exc_tb = sys.exc_info()
            self.log.error('Unable to obtain mac_address, %s' % exc_type)
            self.log.error('  exception traceback info: %s' % exc_tb)
            mac_address = 'AB:12:CD:34:EF:56:BO:G0:US'

        conn_user_id = None
        try:
            conn_user_id = getpass.getuser()
            self.log.debug('Obtained conn_user_id as: %s' % conn_user_id)
        except:
            self.log.error('Unable to obtain conn_user_id, %s' % sys.exc_info()[0])
            conn_user_id = 'UnobtainableUserID'

        app_base_dir = None
        try:
            app_base_dir = os.getcwd()
            self.log.debug('Obtained app_base_dir as: %s' % app_base_dir)
        except:
            self.log.error('Unable to obtain app_base_dir, %s' % sys.exc_info()[0])
            app_base_dir = 'UnobtainableApplicationDirectoryPath'

        # app_crtime    = str(os.path.getctime(app_base_dir))  # this is unreliable

        config_dir = None
        try:
            config_dir = os.path.split(config_path)[0]
            self.log.debug('Obtained config_dir as: %s' % config_dir)
        except:
            self.log.error('Unable to obtain config_dir, %s' % sys.exc_info()[0])
            config_dir = 'UnobtainableConfigDirPath'

        app_owner_id = None
        try:
            app_owner_id = self.getOwner(operating_system)
            self.log.debug('Obtained app_owner_id as: %s' % app_owner_id)
        except:
            self.log.error('Unable to obtain app_owner_id, %s' % sys.exc_info()[0])
            app_owner_id = 'UnobtainableApplicationOwnerID'

        conf_owner_id = None
        try:
            conf_owner_id = self.getOwner(operating_system, config_path)
            self.log.debug('Obtained conf_owner_id as: %s' % conf_owner_id)
        except:
            self.log.error('Unable to obtain conf_owner_id, %s' % sys.exc_info()[0])
            conf_owner_id = 'UnobtainableConfigOwnerID'

        config_name = None
        try:
            config_name = os.path.basename(config_path)
            self.log.debug('Obtained config_name as: %s' % config_name)
        except:
            self.log.error('Unable to obtain config_name, %s' % sys.exc_info()[0])
            config_name = 'LostConfigNameAlert'

        ac            = ac_conf['Server']
        workspace     = ac_conf['Workspace']
        project       = ac_conf['Project']

        items = [hostname, hardware_id, mac_address, conn_user_id, app_base_dir, app_owner_id,
                     config_dir, config_name, conf_owner_id, ac, workspace, project]

        key_phrase = "-".join(items)
        #print (repr(items))
        return key_phrase
